/* Snark - Main snark program startup class.
   Copyright (C) 2003 Mark J. Wielaard

   This file is part of Snark.
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

package org.klomp.snark;

import java.io.*;
import java.net.*;
import java.util.*;

import org.klomp.snark.bencode.*;

/**
 * Main Snark program startup class.
 *
 * @author Mark Wielaard (mark@klomp.org)
 */
public class Snark
{
  private final static int MIN_PORT = 6881;
  private final static int MAX_PORT = 6889;

  /**
   * Whether or not the show debug info.
   */
  public static boolean debug;

  /**
   * If not -1 the port number to listen on.
   * Otherwise it will be the first freee port between MIN_PORT and MAX_PORT.
   */
  private static int user_port = -1;

  /**
   * If non-null the ip address or hostname to run the tracker on.
   */
  private static String ip;

  /**
   * Either a local torrent file, a URL pointing to a torrent or a
   * file or directory to create a torrent for.
   */
  private static String torrent;

  public static void main(String[] args)
  {
    // Parse debug, share/ip and torrent file options.
    parseArguments(args);

    // "Taking Three as the subject to reason about--
    // A convenient number to state--
    // We add Seven, and Ten, and then multiply out
    // By One Thousand diminished by Eight.
    //
    // "The result we proceed to divide, as you see,
    // By Nine Hundred and Ninety Two:
    // Then subtract Seventeen, and the answer must be
    // Exactly and perfectly true.

    // Create a new ID and fill it with something random.  First nine
    // zeros bytes, then three bytes filled with three and then
    // sixteen random bytes.
    byte[] id = new byte[20];
    Random random = new Random();
    int i;
    for (i = 0; i < 9; i++)
      id[i] = 0;
    id[i++] = 3;
    id[i++] = 3;
    id[i++] = 3;
    while (i < 20)
      id[i++] = (byte)random.nextInt(256);

    int port;
    IOException lastException = null;
    ServerSocket serversocket = null;
    if (user_port != -1)
      {
	port = user_port;
	try
	  {
	    serversocket = new ServerSocket(port);
	  }
	catch(IOException ioe)
	  {
	    lastException = ioe;
	  }
      }
    else
      {
	for (port = MIN_PORT; serversocket == null && port <= MAX_PORT; port++)
	  {
	    try
	      {
		serversocket = new ServerSocket(port);
	      }
	    catch(IOException ioe)
	      {
		lastException = ioe;
	      }
	  }
      }
    if (serversocket == null)
      {
	String message = "Cannot accept incoming connections ";
	if (user_port == -1)
	  message = message + "tried ports " + MIN_PORT + " - " + MAX_PORT;
	else
	  message = message + "on port " + user_port ;
	
	if (ip != null || user_port != -1)
	  fatal(message, lastException);
	else
	  System.out.println("WARNING: " + message);
	port = -1;
      }
    else
      {
	port = serversocket.getLocalPort();
	System.out.println("Listening on port: " + port);
      }

    // Figure out what the torrent argument represents.
    MetaInfo meta = null;
    Storage storage = null;
    File f = null;
    try
      {
	InputStream in;
	f = new File(torrent);
	if (f.exists())
	  in = new FileInputStream(f);
	else
	  {
	    URL u = new URL(torrent);
	    URLConnection c = u.openConnection();
	    c.connect();
	    in = c.getInputStream();

	    if (c instanceof HttpURLConnection)
	      {
		// Check whether the page exists
		int code = ((HttpURLConnection)c).getResponseCode();
		if (code/100 != 2) // We can only handle 200 OK responses
		  fatal("Loading page '" + args[0] + "' gave error code "
			+ code + ", it probably doesn't exists");
	      }
	  }
	meta = new MetaInfo(new BDecoder(in));
      }
    catch(IOException ioe)
      {
	// OK, so it wasn't a torrent metainfo file.
	if (f != null && f.exists())
	  if (ip == null)
	    fatal("'" + torrent + "' exists,"
		  + " but is not a valid torrent metainfo file."
		  + System.getProperty("line.separator")
		  + "  (use --share to create a torrent from it"
		  + " and start sharing)", ioe);
		 else
	    {
	      // Try to create a new metainfo file
	     System.out.println
	       ("Trying to create metainfo torrent for '" + torrent + "'");
	     try
	       {
		 storage = new Storage
		   (f, "http://" + ip + ":" + port + "/announce");
		 meta = storage.getMetaInfo();
	       }
	     catch (IOException ioe2)
	       {
		 fatal("Could not create torrent for '" + torrent + "'", ioe2);
	       }
	    }
	else
	  fatal("Cannot open '" + torrent + "'", ioe);
      }
    
    debug(meta.toString());

    // When the metainfo torrent was created from an existing file/dir
    // it alreadt exists.
    if (storage == null)
      {
	try
	  {
	    storage = new Storage(meta);
	  }
	catch (IOException ioe)
	  {
	    fatal("Could not create storage", ioe);
	  }
      }

    PeerCoordinator coordinator = new PeerCoordinator(id, meta, storage);
    HttpAcceptor httpacceptor;
    if (ip != null)
      {
	MetaInfo m = meta.reannounce
	  ("http://" + ip + ":" + port + "/announce");
	Tracker tracker = new Tracker(m);
	try
	  {
	    tracker.addPeer(new PeerID(id, InetAddress.getByName(ip), port));
	  }
	catch (UnknownHostException oops)
	  {
	    fatal("Could not start tracker for " + ip, oops);
	  }
	httpacceptor = new HttpAcceptor(tracker);
      }
    else
      httpacceptor = null;

    PeerAcceptor peeracceptor = new PeerAcceptor(coordinator);
    ConnectionAcceptor acceptor = new ConnectionAcceptor(serversocket,
							 httpacceptor,
							 peeracceptor);

    if (ip != null)
      System.out.println("Torrent available on "
			 + "http://" + ip + ":" + port
			 + "/metainfo.torrent");

    TrackerClient trackerclient = new TrackerClient(meta, coordinator, port);
    trackerclient.start();

    SnarkShutdown snarkhook = new SnarkShutdown(trackerclient);
    Runtime.getRuntime().addShutdownHook(snarkhook);
  }

  /**
   * Sets debug, ip and torrent variables.
   * Calls usage(), which terminates the program, if non-valid argument list.
   */
  private static void parseArguments(String[] args)
  {
    int i = 0;
    while (i < args.length)
      {
	if (args[i].equals("--debug"))
	  {
	    debug = true;
	    i++;
	  }
	else if (args[i].equals("--port"))
	  {
	    if (args.length - 1 < i + 1)
	      usage("--port needs port number to listen on");
	    try
	      {
		user_port = Integer.parseInt(args[i + 1]);
	      }
	    catch (NumberFormatException nfe)
	      {
		usage("--port argument must be a number (" + nfe + ")");
	      }
	    i += 2;
	  }
	else if (args[i].equals("--share"))
	  {
	    if (args.length - 1 < i + 1)
	      usage("--share needs local ip-address or host-name");
	    ip = args[i + 1];
	    i += 2;
	  }
	else
	  {
	    torrent = args[i];
	    i++;
	    break;
	  }
      }

    if (torrent == null || i != args.length)
      if (torrent != null && torrent.startsWith("-"))
	usage("Unknow option '" + torrent + "'.");
      else
	usage("Need exactly one <url>, <file> or <dir>.");
  }
  
  private static void usage(String s)
  {
    System.out.println("snark: " + s);
    usage();
  }

  private static void usage()
  {
    System.out.println
      ("Usage: snark [--debug] [--port <port>] [--share (<ip>|<host>)]");
    System.out.println
      ("            <url>|<file>|<dir>)");
    System.out.println
      ("  --debug\tShows extra info and stacktraces.");
    System.out.println
      ("  --port\tThe port to listen on for incomming connections");
    System.out.println
      ("        \t(if not given defaults to first free port between "
       + MIN_PORT + "-" + MAX_PORT + ").");
    System.out.println
      ("  --share\tStart torrent tracker on <ip> address or <host> name.");
    System.out.println
      ("  <url>  \tURL pointing to .torrent metainfo file to download/share.");
    System.out.println
      ("  <file> \tEither a local .torrent metainfo file to download");
    System.out.println
      ("         \tor (with --share) a file to share.");
    System.out.println
      ("  <dir>  \tA directory with files to share (needs --share).");
    System.exit(-1);
  }

  /**
   * Aborts program abnormally.
   */
  public static void fatal(String s)
  {
    fatal(s, null);
  }

  /**
   * Aborts program abnormally.
   */
  public static void fatal(String s, Throwable t)
  {
    System.err.println("snark: " + s
		       + ((debug || t == null) ? "" : (": " + t)));
    if (debug && t != null)
      t.printStackTrace();
    System.exit(-1);
  }

  /**
   * Show debug info if debug is true.
   */
  public static void debug(String s)
  {
    if (debug)
      System.out.println(s);
  }
}
